/* ****************************************************** */
/* nesting.c                                              */
/* Count brace nesting in TeX files                       */
/*                                                        */
/* Author:    vodoo@vakw.ch                               */
/* Copyright: 2012 vodoo@vakw.ch                          */
/* License:   GPLv2                                       */
/* ****************************************************** */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define  BUFSIZE 4096
unsigned char linebuf[BUFSIZE+1];

int lbp      = 0;  /* pointer into buffer         */
int escape   = 0;  /* escape mode on/off switch   */
int comment  = 0;  /* comment mode switch         */
int premath  = 0;  /* single dollar sign flag     */
int mathmode = 0;  /* math mode on/off switch     */


int main(int argc, char *argv[]) {

  FILE *infile;
  int c;           /* input character */
  int nl = 0;      /* nesting level   */
  int cleanup = 0; /* cleanup switch  */

  /* Check for required file argument */
  if(argc < 2) {
    fprintf(stderr, "%s: %s\n%s %s %s\n", argv[0], "Missing file argument",
      "Usage:", argv[0], "<filename>");
    exit(1);
  }

  if((argc >= 2) && ((strncmp(argv[1], "-h", 2)        == 0)
                 ||  (strncmp(argv[1], "--help", 6)    == 0)
                 ||  (strncmp(argv[1], "-v", 2)        == 0)
                 ||  (strncmp(argv[1], "--version", 9) == 0))) {
    printf("%s\n", argv[0]);
    printf("Version: 1.0\n");
    printf("Usage:   nesting <filename>\n");
    printf("         nesting -h | --help | -v | --version\n\n");
    printf("Show the contents of the input file on STDOUT, each line preceeded\n");
    printf("by a number indicating the nesting level of braces '{}' at the end\n");
    printf("of the respective line.\n\n");
    printf("TeX syntax is used. Braces in comments are not taken into account.\n");
    return(0);
  }

  /* Open the input file for read */
  if((infile = fopen(argv[1], "r")) == NULL) {
    fprintf(stderr, "%s %s\n", "Error: could not open file", argv[1]);
    exit(1);
  }

  /* Read from file */
  while((c = fgetc(infile)) != EOF) {

    /* Complete dangling mathmode settings for a single $ */
    if(premath && (c != '$') && (! comment)) {
      premath = 0;
      if(mathmode) {
        mathmode = 0;
      } else {
        mathmode = 1;
      }
    }

    /* Count braces but ignore escaped braces in mathmode */
    if(! comment) {
      if(mathmode) {
        if(! escape) {
          if(c == '{') {
            ++nl;
          }
          if(c == '}') {
            --nl;
          }
        }
      } else { /* not in mathmode */
        if(c == '{') {
          ++nl;
        }
        if(c == '}') {
          --nl;
        }
      }
    }

    /* Show counter and input line when reaching the end of the line */
    if(c == '\n') {
      cleanup = 0;
      linebuf[lbp] = '\0';
      lbp = 0;
      comment = 0;
      printf("%3d %s\n", nl, linebuf);
    } else {
      /* store in buffer */
      if(lbp < BUFSIZE) {
        linebuf[lbp++] = (unsigned char) c;
        cleanup = 1;
      } else {
        linebuf[lbp] = '\0';
        printf("%3d %s\n", nl, linebuf);
        fflush(stdout);
        fclose(infile);
        fprintf(stderr, "%s\n", "Error: line too long; buffer space is full");
        exit(1);
      }
    }

    /* % starts a comment unless preceded by an escape character */
    if(c == '%') {
      if(! comment) {
        if(! escape) {
          comment = 1;
        }
      }
    }

    /* \ starts escaping but is ignored in comments */
    if(c == '\\') {
      if(! comment) {
        escape = 1;
      }
    } else { /* any other character resets escape mode */
      escape = 0;
    }

    /* Math mode starts and ends with $ or $$ */
    if((c == '$') && (! comment)) {
      if(mathmode) {
        if(premath) {
          mathmode = 0;
          premath  = 0;
        } else {
          premath =  1;
        }
      } else {
        if(premath) {
          mathmode = 1;
          premath  = 0;
        } else {
          premath =  1;
        }
      }
    }
  }

  fclose(infile);

  /* print unfinished line (there is no \nl at the end of the file) */
  if(cleanup) {
    linebuf[lbp] = '\0';
    printf("%3d %s\n", nl, linebuf);
  }

  fflush(stdout);

  if(mathmode) {
    fprintf(stderr, "%s\n", "Warning: file ended in math mode");
  }

  if(nl > 0 || nl < 0) {
    fprintf(stderr, "%s %d\n", 
      "Warning: file ended with brace nesting of level", nl);
  }

  fflush(stderr);
  return(0);
}
